/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.binding.soap.listener.incoming;

import static org.ow2.petals.binding.soap.Constants.HttpServer.DEFAULT_HTTP_ACCEPTORS;
import static org.ow2.petals.binding.soap.Constants.HttpServer.DEFAULT_HTTP_HOST;
import static org.ow2.petals.binding.soap.Constants.HttpServer.DEFAULT_HTTP_PORT;
import static org.ow2.petals.binding.soap.Constants.HttpServer.DEFAULT_HTTP_THREAD_POOL_SIZE_MAX;
import static org.ow2.petals.binding.soap.Constants.HttpServer.DEFAULT_HTTP_THREAD_POOL_SIZE_MIN;

import java.net.Inet4Address;
import java.net.InetAddress;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.ow2.petals.binding.soap.util.NetworkUtil;

/**
 * This class is used to store only the SOAP server related configuration
 * values.
 * 
 * @author Christophe HAMERLING - eBMWebSourcing
 * 
 */
public class SoapServerConfig {

	/**
	 * The port used for incoming SOAP requests
	 */
	private int port = DEFAULT_HTTP_PORT;

	/**
	 * The current interfaces which can be use for incoming listening
	 */
	private final List<InetAddress> addresses;

	/**
	 * The host
	 */
	private String host;

	/**
	 * Restrict access to a particular network interface
	 */
	private boolean restrict = false;

	/**
	 * The servlet mappping : axis2/<b>services</b>
	 */
	private String servicesMapping;

	/**
	 * The servlet context : <b>axis2</b>/services
	 */
	private String servicesContext;

	/**
	 * The protocol used for this server (http / https ...)
	 */
	private String protocol;

	/**
	 * The server provides/!provides the services list
	 */
	private boolean providesList;
	
	private HashMap<String, String> redirects = null;

	/**
	 * Default value is set to 255 (The jetty implementation default)
	 */
	protected int jettyThreadMaxPoolSize = DEFAULT_HTTP_THREAD_POOL_SIZE_MAX;

	/**
	 * Default value is set to 1 (The jetty implementation default)
	 */
	protected int jettyThreadMinPoolSize = DEFAULT_HTTP_THREAD_POOL_SIZE_MIN;

	/**
	 * Default value is set to 4
	 */
	protected int jettyAcceptors = DEFAULT_HTTP_ACCEPTORS;

	protected String wsnContext = "wsn";

	protected String wsnProducerMapping = "producer";

	protected String baseURL;

	protected String wsnProducerURL;

	protected String servicesURL;

	/**
	 * Creates a new instance of {@link SoapServerConfig}
	 * 
	 */
	public SoapServerConfig() {
		addresses = new ArrayList<InetAddress>();
	}

	/**
	 * @param host
	 *            the host to set
	 */
	protected void addAddress(final InetAddress ia) {
		addresses.add(ia);
	}

	/**
	 * Add all the {@link InetAddress}
	 * 
	 * @param addresses
	 */
	protected void addAddresses(final Set<Inet4Address> addresses) {
		this.addresses.addAll(addresses);
	}

	/**
	 * Get the root URL
	 * 
	 * @return
	 */
	private String buildBaseURL() {
		final StringBuffer sb = new StringBuffer(getProtocol());
		sb.append("://");
		sb.append(getHost());
		if (getPort() > 0) {
			sb.append(':');
			sb.append(getPort());
		}

		return sb.toString();
	}

	/**
	 * Get the host. If restricted, get the first address of the addresses set.
	 * If not restricted, try to get an address which is not the local one since
	 * it can be used from external callers.
	 * 
	 * @return
	 */
	public String buildHost() {
		InetAddress result = null;

		if (restrict) {
			// get the first host, no matter if it is the loopback since it has
			// been configured by the user to be restricted ! The set should
			// contains only the user address
			if (addresses != null && addresses.size() > 0) {
				result = addresses.get(0);
			}
		} else {
			// get the first inet address which is not the loopback one since it
			// can be called from external clients
			if (addresses != null && addresses.size() > 0) {
				final Iterator<InetAddress> iter = addresses.iterator();
				while (iter.hasNext() && result == null) {
					final InetAddress addr = iter.next();
					if (!NetworkUtil.isLoopbackAddress(addr)) {
						result = addr;
					}
				}
			}
		}
		return (result == null ? DEFAULT_HTTP_HOST : result.getHostAddress());
	}

	// /**
	// * @return the host
	// */
	// public InetAddress getHost() {
	// return this.host;
	// }

	/**
	 * Build the services URL from the current configuration properties
	 * 
	 * @return
	 */
	private String buildServicesURL() {
		final StringBuffer sb = new StringBuffer(getBaseURL());
		sb.append('/');
		sb.append(getServicesContext());
		sb.append('/');
		sb.append(getServicesMapping());
		return sb.toString();
	}

	/**
	 * 
	 * @return
	 */
	public String getBaseURL() {
		if (baseURL == null) {
			baseURL = buildBaseURL();
		}
		return baseURL;
	}

	/**
	 * 
	 * @return
	 */
	public String getHost() {
		if (host == null) {
			host = buildHost();
		}
		return host;
	}

	/**
	 * @return the jettyAcceptors
	 */
	public int getJettyAcceptors() {
		return jettyAcceptors;
	}

	/**
	 * @return the jettyThreadMaxPoolSize
	 */
	public int getJettyThreadMaxPoolSize() {
		return jettyThreadMaxPoolSize;
	}

	/**
	 * @return the jettyThreadMinPoolSize
	 */
	public int getJettyThreadMinPoolSize() {
		return jettyThreadMinPoolSize;
	}

	/**
	 * @return the port
	 */
	public int getPort() {
		return port;
	}

	/**
	 * @return the protocol
	 */
	public String getProtocol() {
		return protocol;
	}

	/**
	 * @return the servicesContext
	 */
	public String getServicesContext() {
		return servicesContext;
	}

	/**
	 * @return the servicesMapping
	 */
	public String getServicesMapping() {
		return servicesMapping;
	}

	/**
	 * Get the services URL
	 * 
	 * @return
	 */
	public String getServicesURL() {
		if (servicesURL == null) {
			servicesURL = buildServicesURL();
		}
		return servicesURL;
	}

	/**
	 * @return the providesList
	 */
	public boolean isProvidesList() {
		return providesList;
	}

	/**
	 * @return the restrict
	 */
	public boolean isRestrict() {
		return restrict;
	}

	/**
	 * @param jettyAcceptors
	 *            the jettyAcceptors to set
	 */
	public void setJettyAcceptors(final int jettyAcceptors) {
		this.jettyAcceptors = jettyAcceptors;
	}

	/**
	 * @param jettyThreadMaxPoolSize
	 *            the jettyThreadMaxPoolSize to set
	 */
	public void setJettyThreadMaxPoolSize(final int jettyThreadMaxPoolSize) {
		this.jettyThreadMaxPoolSize = jettyThreadMaxPoolSize;
	}

	/**
	 * @param jettyThreadMinPoolSize
	 *            the jettyThreadMinPoolSize to set
	 */
	public void setJettyThreadMinPoolSize(final int jettyThreadMinPoolSize) {
		this.jettyThreadMinPoolSize = jettyThreadMinPoolSize;
	}

	/**
	 * @param port
	 *            the port to set
	 */
	public void setPort(final int port) {
		this.port = port;
	}

	/**
	 * @param protocol
	 *            the protocol to set
	 */
	public void setProtocol(final String protocol) {
		this.protocol = protocol;
	}

	/**
	 * @param providesList
	 *            the providesList to set
	 */
	public void setProvidesList(final boolean providesList) {
		this.providesList = providesList;
	}

	/**
	 * @param restrict
	 *            the restrict to set
	 */
	public void setRestrict(final boolean restrict) {
		this.restrict = restrict;
	}

	/**
	 * @param servicesContext
	 *            the servicesContext to set
	 */
	public void setServicesContext(final String servicesContext) {
		this.servicesContext = servicesContext;
	}

	/**
	 * @param servicesMapping
	 *            the servicesMapping to set
	 */
	public void setServicesMapping(final String servicesMapping) {
		this.servicesMapping = servicesMapping;
	}
	
	/**
	 * Define a HTTP redirection (to customize URLs)
	 * @param from Original URI
	 * @param to Destination URI
	 */
	public void addRedirect(String from, String to) {
		if(redirects == null) redirects = new HashMap<String, String>();
		redirects.put(from, to);
	}
	public String getRedirect(String from) {
		if(redirects == null) return null;
		else return redirects.get(from);
	}
	public void removeRedirect(String from) {
		if(redirects != null) redirects.remove(from);
	}
}
